#pragma once

#include <array>
#include <tuple>
#include <QObject>
#include <bgtu/computer_graphics_lab_work/renderer_api/sprite.hpp>
#include <bgtu/computer_graphics_lab_work/utilities/zoomed_voxel_painter.hpp>
#include "sprite_data.hpp"

namespace BGTU::ComputerGraphicsLabWork::Lab1 {
	struct variant_sprites {
	private :
		using sprite_t = RendererApi::Sprite<Lab1SpriteData::ShapeData, Utilities::ZoomedVoxelPainter<Utilities::DefaultVoxelDrawerCache::VoxelPainterImpl>>;
	public:


		sprite_t *const *const sprites;
		std::size_t const count;


		constexpr variant_sprites(sprite_t **_sprites, std::size_t _count) : sprites{_sprites}, count{_count} {

		}

	private:
		template<class ...sprites_t>
		struct _make {
		};

		template<>
		struct _make<> {
		public:
			constexpr _make() {};

			constexpr void export_pointers(sprite_t **dst, std::size_t pos) {
			}
		};

		template<class e0>
		struct _make<e0> {
		public:
			e0 value;

			constexpr _make(e0 v) : value{std::move(v)} {};

			constexpr void export_pointers(sprite_t **dst, std::size_t pos) {
				dst[pos] = &this->value;
			}
		};

		template<class e0, class ...en>
		struct _make<e0, en...> {
		public:
			e0 value;
			_make<en...> next;

			constexpr _make(e0 v, en...nv) : value{std::move(v)}, next{nv...} {};

			constexpr void export_pointers(sprite_t **dst, std::size_t pos) {
				dst[pos] = &this->value;
				this->next.export_pointers(dst, pos + 1);
			}
		};

	public:

		template<class ...sprites_t>
		constexpr static variant_sprites make(sprites_t...sprites) {
			static _make<sprites_t...> i{sprites...};
			static sprite_t *p[sizeof...(sprites)];

			i.export_pointers(p, 0);
			return variant_sprites{p, sizeof...(sprites)};
		}

	private:

		template<void (...sprites)(Utilities::ZoomedVoxelPainter<BGTU::ComputerGraphicsLabWork::Utilities::DefaultVoxelDrawerCache::VoxelPainterImpl> *frame,
								   Lab1SpriteData::ShapeData const *data)>
		struct _make_light : public RendererApi::Sprite<Lab1SpriteData::ShapeData, Utilities::ZoomedVoxelPainter<Utilities::DefaultVoxelDrawerCache::VoxelPainterImpl>> {
		};

		template<>
		struct _make_light<> {
		public:
			constexpr _make_light() {};

			constexpr void export_pointers(sprite_t **dst, std::size_t pos) {
			}
		};

		template<void (s0)(Utilities::ZoomedVoxelPainter<BGTU::ComputerGraphicsLabWork::Utilities::DefaultVoxelDrawerCache::VoxelPainterImpl> *frame,
						   Lab1SpriteData::ShapeData const *data),
				void (...sn)(Utilities::ZoomedVoxelPainter<BGTU::ComputerGraphicsLabWork::Utilities::DefaultVoxelDrawerCache::VoxelPainterImpl> *frame,
							 Lab1SpriteData::ShapeData const *data)>
		struct _make_light<s0, sn...> : public RendererApi::Sprite<Lab1SpriteData::ShapeData, Utilities::ZoomedVoxelPainter<Utilities::DefaultVoxelDrawerCache::VoxelPainterImpl>> {
		public:
			_make_light<sn...> next;

			constexpr _make_light() : next{} {};

			constexpr void export_pointers(sprite_t **dst, std::size_t pos) {
				dst[pos] = this;
				this->next.export_pointers(dst, pos + 1);
			}

			void draw(Utilities::ZoomedVoxelPainter<BGTU::ComputerGraphicsLabWork::Utilities::DefaultVoxelDrawerCache::VoxelPainterImpl> *frame, const Lab1::Lab1SpriteData::ShapeData *data) const final {
				s0(frame, data);
			}

			void clicked() final {}
		};


	public:
		template<void (...sprites)(Utilities::ZoomedVoxelPainter<BGTU::ComputerGraphicsLabWork::Utilities::DefaultVoxelDrawerCache::VoxelPainterImpl> *frame,
								   Lab1SpriteData::ShapeData const *data)>
		constexpr static variant_sprites make_light() {
			static _make_light<sprites...> i{};
			static sprite_t *p[sizeof...(sprites)];

			i.export_pointers(p, 0);
			return variant_sprites{p, sizeof...(sprites)};
		}
	};


	extern variant_sprites variant1;
	extern variant_sprites variant2;
	extern variant_sprites variant3;
	extern variant_sprites variant4;
	extern variant_sprites variant5;
	extern variant_sprites variant6;
	extern variant_sprites variant8;
	extern variant_sprites variant9;

	class VariantsManager : public QObject {
	Q_OBJECT
	signals:

		void set_sprites(
				RendererApi::Sprite<Lab1SpriteData::ShapeData, Utilities::ZoomedVoxelPainter<Utilities::DefaultVoxelDrawerCache::VoxelPainterImpl>> *const *sprites,
				std::size_t count
				);

	private:

		inline void _set_sprites(variant_sprites sprites) {
			emit this->set_sprites(sprites.sprites, sprites.count);
		}

	public slots:

		inline void set_variant_1() {
			this->_set_sprites(variant1);
		};

		inline void set_variant_2() {
			this->_set_sprites(variant2);
		};

		inline void set_variant_3() {
			this->_set_sprites(variant3);
		};

		inline void set_variant_4() {
			this->_set_sprites(variant4);
		};

		inline void set_variant_5() {
			this->_set_sprites(variant5);
		};

		inline void set_variant_6() {
			this->_set_sprites(variant6);
		};

		inline void set_variant_8() {
			this->_set_sprites(variant8);
		};

		inline void set_variant_9() {
			this->_set_sprites(variant9);
		};
	};
}