#pragma once

#include <cmath>
#include <bgtu/computer_graphics_lab_work/renderer_api/color.hpp>
#include <bgtu/computer_graphics_lab_work/renderer_api/point.hpp>
#include <bgtu/computer_graphics_lab_work/utilities/shapes/circle.hpp>
#include <bgtu/computer_graphics_lab_work/utilities/shapes/polygon.hpp>
#include "../sprite_data.hpp"

namespace BGTU::ComputerGraphicsLabWork::Lab1 {
	template<double z, RendererApi::Color::Transparent c, double radius_multiplier>
	void static_centered_circle_edge(Utilities::ZoomedVoxelPainter<BGTU::ComputerGraphicsLabWork::Utilities::DefaultVoxelDrawerCache::VoxelPainterImpl> *frame, Lab1SpriteData::ShapeData const *data) {
		Utilities::Shapes::draw_circle_edge(frame, {0, 0}, data->radius * radius_multiplier, z, c);
	}

	template<RendererApi::PointF<2>... points_m>
	struct _iter_points {
	};

	template<>
	struct _iter_points<> {
	public:
		static void apply_radius(RendererApi::PointI<2> *dst, std::size_t i, RendererApi::PointI<2>::component_t radius) {}


		static void as_radius_and_angle_pos(RendererApi::PointI<2> *dst, std::size_t i, Lab1SpriteData::ShapeData const *data) {}

		static void as_radius_and_angle_neg(RendererApi::PointI<2> *dst, std::size_t i, Lab1SpriteData::ShapeData const *data) {}

		static void as_radius_and_angle_static(RendererApi::PointI<2> *dst, std::size_t i, Lab1SpriteData::ShapeData const *data) {}
	};

	template<RendererApi::PointF<2> p0, RendererApi::PointF<2>... pn>
	struct _iter_points<p0, pn...> {
	public:
		static void apply_radius(RendererApi::PointI<2> *dst, std::size_t i, RendererApi::PointI<2>::component_t radius) {
			dst[i] = {(RendererApi::PointI<2>::component_t) (p0.x * radius), (RendererApi::PointI<2>::component_t) (p0.y * radius)};
			_iter_points<pn...>::apply_radius(dst, i + 1, radius);
		}

		static void as_radius_and_angle_pos(RendererApi::PointI<2> *dst, std::size_t i, Lab1SpriteData::ShapeData const *data) {
			dst[i] = data->pos_rotated(data->radius * p0.x, p0.y);
			_iter_points<pn...>::as_radius_and_angle_pos(dst, i + 1, data);
		}

		static void as_radius_and_angle_neg(RendererApi::PointI<2> *dst, std::size_t i, Lab1SpriteData::ShapeData const *data) {
			dst[i] = data->neg_rotated(data->radius * p0.x, p0.y);
			_iter_points<pn...>::as_radius_and_angle_neg(dst, i + 1, data);
		}

		static void as_radius_and_angle_static(RendererApi::PointI<2> *dst, std::size_t i, Lab1SpriteData::ShapeData const *data) {
			dst[i] = data->static_rotated(data->radius * p0.x, p0.y);
			_iter_points<pn...>::as_radius_and_angle_static(dst, i + 1, data);
		}
	};

	template<double z, RendererApi::Color::Transparent c, RendererApi::PointF<2>... points_m>
	void static_polygon_edge(Utilities::ZoomedVoxelPainter<BGTU::ComputerGraphicsLabWork::Utilities::DefaultVoxelDrawerCache::VoxelPainterImpl> *frame, Lab1SpriteData::ShapeData const *data) {
		union {
			RendererApi::PointI<2> array[sizeof...(points_m)];
			bool nothing;
		} points{.nothing = false};
		_iter_points<points_m...>::apply_radius(points.array, 0, data->radius);
		Utilities::Shapes::draw_polygon_edge(frame, std::move(points.array), z, c);
	}

	template<double z, RendererApi::Color::Transparent c, RendererApi::PointF<2>... points_m>
	void pos_rotated_polygon_edge(Utilities::ZoomedVoxelPainter<BGTU::ComputerGraphicsLabWork::Utilities::DefaultVoxelDrawerCache::VoxelPainterImpl> *frame, Lab1SpriteData::ShapeData const *data) {
		union {
			RendererApi::PointI<2> array[sizeof...(points_m)];
			bool nothing;
		} points{.nothing = false};
		_iter_points<points_m...>::as_radius_and_angle_pos(points.array, 0, data);
		Utilities::Shapes::draw_polygon_edge(frame, std::move(points.array), z, c);
	}

	template<double z, RendererApi::Color::Transparent c, RendererApi::PointF<2>... points_m>
	void neg_rotated_polygon_edge(Utilities::ZoomedVoxelPainter<BGTU::ComputerGraphicsLabWork::Utilities::DefaultVoxelDrawerCache::VoxelPainterImpl> *frame, Lab1SpriteData::ShapeData const *data) {
		union {
			RendererApi::PointI<2> array[sizeof...(points_m)];
			bool nothing;
		} points{.nothing = false};
		_iter_points<points_m...>::as_radius_and_angle_neg(points.array, 0, data);
		Utilities::Shapes::draw_polygon_edge(frame, std::move(points.array), z, c);
	}

	template<double z, RendererApi::Color::Transparent c, RendererApi::PointF<2>... points_m>
	void static_rotated_polygon_edge(Utilities::ZoomedVoxelPainter<BGTU::ComputerGraphicsLabWork::Utilities::DefaultVoxelDrawerCache::VoxelPainterImpl> *frame, Lab1SpriteData::ShapeData const *data) {
		union {
			RendererApi::PointI<2> array[sizeof...(points_m)];
			bool nothing;
		} points{.nothing = false};
		_iter_points<points_m...>::as_radius_and_angle_static(points.array, 0, data);
		Utilities::Shapes::draw_polygon_edge(frame, std::move(points.array), z, c);
	}
}