#pragma once

#include <utility>
#include <algorithm>
#include <Qt>
#include <QDebug>
#include <QObject>
#include <QThread>
#include <QMutex>
#include <QElapsedTimer>
#include <bgtu/computer_graphics_lab_work/renderer_api/voxel_painter.hpp>
#include <bgtu/computer_graphics_lab_work/renderer_api/sprite.hpp>
#include <bgtu/computer_graphics_lab_work/utilities/voxel_painter_exporter_to_packed_array.hpp>
#include "owned_qimage.hpp"

namespace BGTU::ComputerGraphicsLabWork::QtUtilities {
	template<class sprite_data_t, class renderer_voxel_painter_t, class renderer_context_t>
	using _SeparateThreadedRenderer_RenderProcedureType = void (renderer_context_t::*)(
			RendererApi::VirtualVoxelPainter::visible_pixel_coordinate_fast_t,
			RendererApi::VirtualVoxelPainter::visible_pixel_coordinate_fast_t,
			Utilities::VoxelPainterExporterToBgr32Array *,
			RendererApi::Color,
			RendererApi::Sprite<sprite_data_t, renderer_voxel_painter_t> *const *,
			std::size_t,
			sprite_data_t const *
	);

	template<class sprite_data_t, class renderer_voxel_painter_t, class renderer_context_t, _SeparateThreadedRenderer_RenderProcedureType<sprite_data_t, renderer_voxel_painter_t, renderer_context_t>>
	class SeparateThreadedRenderer;

	class _SeparateThreadedRenderer_Signals : public QObject {
	Q_OBJECT

		template<class sprite_data_t, class renderer_voxel_painter_t, class renderer_context_t, _SeparateThreadedRenderer_RenderProcedureType<sprite_data_t, renderer_voxel_painter_t, renderer_context_t>>
		friend
		class SeparateThreadedRenderer;

	private:
		explicit _SeparateThreadedRenderer_Signals(QObject *owner = nullptr) : QObject{owner} {}

	signals:

		void frame_rendered(OwnedQImage *img);

	public slots:

		virtual void set_frame_size(RendererApi::VirtualVoxelPainter::visible_pixel_coordinate_fast_t width, RendererApi::VirtualVoxelPainter::visible_pixel_coordinate_fast_t height) = 0;

		virtual void set_ms_per_frame(std::uint_fast64_t ms) = 0;

		virtual void set_background(RendererApi::Color) = 0;
	};
}