#include <varargs.h>
#include "frame.hpp"
#include "drawing_assets.hpp"

void draw_circle(Frame *frame, int cx, int cy, int radius, COLOR outline) {
    int x = 0, y = radius;
    while (x < y) {
        int D1 = x * x + y * y - radius * radius;
        int D2 = x * x + (y - 1) * (y - 1) - radius * radius;

        if (D1 > -D2)
            y--;

        frame->set_pixel(cx + x, cy + y, outline);
        frame->set_pixel(cx + x, cy - y, outline);
        frame->set_pixel(cx + y, cy + x, outline);
        frame->set_pixel(cx + y, cy - x, outline);
        frame->set_pixel(cx - x, cy + y, outline);
        frame->set_pixel(cx - x, cy - y, outline);
        frame->set_pixel(cx - y, cy + x, outline);
        frame->set_pixel(cx - y, cy - x, outline);
        x++;
    }
}

void draw_line(Frame *frame, int x1, int y1, int x2, int y2, COLOR color) {
    int dy = y2 - y1, dx = x2 - x1;
    if (dx == 0 && dy == 0) {
        frame->set_pixel(x1, y1, color);
        return;
    }

    if (abs(dx) > abs(dy)) {
        if (x2 < x1) {
            std::swap(x1, x2);
            std::swap(y1, y2);
            dx = -dx;
            dy = -dy;
        }

        int y, dx2 = dx / 2, p = 0;
        if (dy < 0) dx2 = -dx2;
        for (int x = x1; x <= x2; x++) {
            y = (p + dx2) / dx + y1;
            p += dy;
            frame->set_pixel(x, y, color);
        }
    } else {
        if (y2 < y1) {
            std::swap(x1, x2);
            std::swap(y1, y2);
            dx = -dx;
            dy = -dy;
        }

        int x, dy2 = dy / 2, p = 0;
        if (dx < 0) dy2 = -dy2;
        for (int y = y1; y <= y2; y++) {
            x = (p + dy2) / dy + x1;
            p += dx;
            frame->set_pixel(x, y, color);
        }
    }
}

void fill_triangle(Frame *f, int x0, int y0, int x1, int y1, int x2, int y2, Shader const *shader) {
// Отсортируем точки таким образом, чтобы выполнилось условие: y0 < y1 < y2
    if (y1 < y0) {
        std::swap(y1, y0);
        std::swap(x1, x0);
    }
    if (y2 < y1) {
        std::swap(y2, y1);
        std::swap(x2, x1);
    }
    if (y1 < y0) {
        std::swap(y1, y0);
        std::swap(x1, x0);
    }
// Определяем номера строк пикселей, в которых располагаются точки треугольника
    int Y0 = (int) (y0 + 0.5f);
    int Y1 = (int) (y1 + 0.5f);
    int Y2 = (int) (y2 + 0.5f);
// Отсечение невидимой части треугольника
    if (Y0 < 0) Y0 = 0;
    else if (Y0 >= f->height) Y0 = f->height;
    if (Y1 < 0) Y1 = 0;
    else if (Y1 >= f->height) Y1 = f->height;
    if (Y2 < 0) Y2 = 0;
    else if (Y2 >= f->height) Y2 = f->height;
// Рисование верхней части треугольника
    for (float y = Y0 + 0.5f; y < Y1; y++) {
        int X0 = (int) ((y - y0) / (y1 - y0) * (x1 - x0) + x0 + 0.5f);
        int X1 = (int) ((y - y0) / (y2 - y0) * (x2 - x0) + x0 + 0.5f);
        if (X0 > X1) std::swap(X0, X1);
        if (X0 < 0) X0 = 0;
        if (X1 > f->width) X1 = f->width;
        for (int x = X0; x < X1; x++) {
            f->set_pixel(x, y, shader->get_color(x + 0.5f, y));
        }
    }
// Рисование нижней части треугольника
    for (float y = Y1 + 0.5f; y < Y2; y++) {
        int X0 = (int) ((y - y1) / (y2 - y1) * (x2 - x1) + x1 + 0.5f);
        int X1 = (int) ((y - y0) / (y2 - y0) * (x2 - x0) + x0 + 0.5f);
        if (X0 > X1) std::swap(X0, X1);
        if (X0 < 0) X0 = 0;
        if (X1 > f->width) X1 = f->width;
        for (int x = X0; x < X1; x++) {
            f->set_pixel(x, y, shader->get_color(x + 0.5f, y));
        }
    }
}



COLOR ColorFromHSV(double hue, double saturation, double value) {
    int hi = int(floor(hue / 60)) % 6;
    double f = hue / 60 - floor(hue / 60);
    value = value * 255;
    const auto v = (unsigned char) (value);
    const auto p = (unsigned char) (value * (1 - saturation));
    const auto q = (unsigned char) (value * (1 - f * saturation));
    const auto t = (unsigned char) (value * (1 - (1 - f) * saturation));
    switch (hi) {
        case 0:
            return COLOR{v, t, p};
        case 1:
            return COLOR{q, v, p};
        case 2:
            return COLOR{p, v, t};
        case 3:
            return COLOR{p, q, v};
        case 4:
            return COLOR{t, p, v};
        default:
            return COLOR{v, p, q};
    }
}
