#pragma once

#include <cmath>
#include <utility>

struct COLOR {
    unsigned char red;
    unsigned char green;
    unsigned char blue;
    unsigned char alpha;

    COLOR() : red(0), green(0), blue(0), alpha(255) {}

    COLOR(unsigned char red, unsigned char green, unsigned char blue, unsigned char alpha = 255) : red(red), green(green), blue(blue), alpha(alpha) {}

};

class Point {
public:
    int x;
    int y;

    inline Point(int x, int y) : x{x}, y{y} {};

    [[nodiscard]] Point rotated_around(int cx, int cy, double angle_radians) const {
        return Point{
                static_cast<int>((this->x - cx) * std::cos(angle_radians) - (this->y - cy) * std::sin(angle_radians) + cx),
                static_cast<int>((this->x - cx) * std::sin(angle_radians) + (this->y - cy) * std::cos(angle_radians) + cy)
        };
    }

    [[nodiscard]] Point rotated_around(Point p, double angle_radians) const {
        return this->rotated_around(p.x, p.y, angle_radians);
    }

    [[nodiscard]] Point moved(int dx, int dy) const {
        return Point{this->x + dx, this->y + dy};
    }
};

class Frame {
    COLOR *pixels;
public:
    const int width, height;

    inline Frame(int _width, int _height) : width{_width}, height{_height} {
        int size = width * height;

        this->pixels = new COLOR[size];
    }

private:
    [[nodiscard]]inline COLOR _get_pixel(int x, int y) const {
        return this->pixels[y * this->width + x];
    }

    [[nodiscard]] inline COLOR &_get_pixel(int x, int y) {
        return this->pixels[y * this->width + x];
    }

public:
    inline void set_pixel(int x, int y, COLOR color) {
        this->_get_pixel(x, y) = color;
    }

    [[nodiscard]] inline COLOR get_pixel(int x, int y) const {
        return this->_get_pixel(x, y);
    }


    inline ~Frame() {
        delete[]this->pixels;
    }
};
