#pragma once

#include <QWidget>
#include <QGridLayout>
#include <QLabel>
#include <QSize>
#include <QResizeEvent>
#include <QMutex>
#include <QMainWindow>


#include "frame.hpp"

class StatusBar : public QWidget {
Q_OBJECT
private:
    QGridLayout *_layout;

    QLabel *_scale_label;
    static const QString _scale_label_template;

    QLabel *_painter_label;
    static const QString _painter_label_template;

    struct {
        QLabel *title;
        QLabel *x;
        QLabel *y;
    } _screen_coordinates, _scaled_coordinates;
    static const QString _x_template, _y_template;


public:
    explicit StatusBar(QWidget *parent);

public slots:

    void set_coordinates(unsigned screen_x, unsigned screen_y, unsigned scaled_x, unsigned scaled_y);

    void set_scale(unsigned current_scale);

    void set_painter_index(unsigned index, unsigned painters_count);
};

class Canvas : public QWidget {
Q_OBJECT
private:
    QMutex _sync;
    QImage _current_frame;

public:
    explicit Canvas(QWidget *parent);

public slots:

    void set_frame(QImage &f);

public:
    void paintEvent(QPaintEvent *event) override;

    void resizeEvent(QResizeEvent *event) override;

signals:

    void resized(QSize new_size);


private:
    struct {
        bool is_pressed;
        QPoint pressed_at;
        bool is_moved_while_pressed;
    } _mouse_processing;
public:

    void keyPressEvent(QKeyEvent *event) override;

    void mousePressEvent(QMouseEvent *event) override;

    void mouseMoveEvent(QMouseEvent *event) override;

    void mouseReleaseEvent(QMouseEvent *event) override;

signals:

    void key_pressed(Qt::Key key);

    void released_mouse_moved(QPoint to);

    void pressed_mouse_moved(QPoint to);

    void mouse_clicked(QPoint where);

    void mouse_dragged(QPoint from, QPoint to);

};

class MainWindow : public QWidget {
Q_OBJECT
private:
    QGridLayout *_layout;
public:
    StatusBar *_status_bar;
    Canvas *_canvas;

    MainWindow();
};