#pragma once


#include <bgtu/computer_graphics_lab_work/renderer_api/color.hpp>
#include "matrix.hpp"


namespace BGTU::ComputerGraphicsLabWork::Utilities {
	class Shader {
	public:
		[[nodiscard]] virtual RendererApi::Color::Transparent get_color(double x, double y) const = 0;

		static Shader const *empty_shader;
	};

	class MonoShader : public Shader {
	private:
		RendererApi::Color::Transparent c;

	public:
		inline explicit MonoShader(RendererApi::Color::Transparent c) : c{c} {}

		[[nodiscard]] inline RendererApi::Color::Transparent get_color(double x, double y) const override {
			return this->c;
		};
	};

	class TransformedShader : public Shader {
	private:
		Shader const *shader;
		Matrix4d transform;

	public:
		inline TransformedShader(Shader const *s, Matrix4d t) : shader{s}, transform{t} {}

		[[nodiscard]] inline RendererApi::Color::Transparent get_color(double x, double y) const override {
			RendererApi::PointF<3> in{x, y, 0};
			RendererApi::PointF<3> out = this->transform * in;
			return this->shader->get_color(out.x, out.y);
		}
	};
}