#pragma once

#include <cstdlib>
#include <bgtu/computer_graphics_lab_work/renderer_api/point.hpp>
#include <bgtu/computer_graphics_lab_work/renderer_api/color.hpp>
#include <bgtu/computer_graphics_lab_work/renderer_api/voxel_painter.hpp>

namespace BGTU::ComputerGraphicsLabWork::Utilities::Shapes {
	template<class receiver_t>
	void iterate_circle_edge(
			RendererApi::PointI<2>::component_t cx, RendererApi::PointI<2>::component_t cy,
			RendererApi::PointI<2>::component_t r,
			receiver_t receiver
	) {
		RendererApi::PointI<2>::component_t x = 0, y = r;
		while (x < y) {
			RendererApi::PointI<2>::component_t D1 = x * x + y * y - r * r;
			RendererApi::PointI<2>::component_t D2 = x * x + (y - 1) * (y - 1) - r * r;

			if (D1 > -D2)
				y--;

			receiver(cx + x, cy + y);
			receiver(cx + x, cy - y);
			receiver(cx + y, cy + x);
			receiver(cx + y, cy - x);
			receiver(cx - x, cy + y);
			receiver(cx - x, cy - y);
			receiver(cx - y, cy + x);
			receiver(cx - y, cy - x);
			x++;
		}
	}


	template<class receiver_t>
	void iterate_circle_edge(
			RendererApi::PointI<2> center,
			RendererApi::PointI<2>::component_t r,
			receiver_t receiver
	) {
		return iterate_circle_edge<receiver_t>(center.x, center.y, r, receiver);
	}


	template<class voxel_painter_t = RendererApi::VirtualVoxelPainter>
	void draw_circle_edge(
			voxel_painter_t *painter,
			RendererApi::PointI<2>::component_t cx, RendererApi::PointI<2>::component_t cy,
			RendererApi::PointI<2>::component_t r,
			double z,
			RendererApi::Color::Transparent outline
	) {
		iterate_circle_edge(
				cx, cy, r,
				[=](RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y) { painter->add_voxel(x, y, z, outline); }
		);
	}

	template<class voxel_painter_t = RendererApi::VirtualVoxelPainter>
	void draw_circle_edge(
			voxel_painter_t *painter,
			RendererApi::PointI<2> center,
			RendererApi::PointI<2>::component_t r,
			double z,
			RendererApi::Color::Transparent outline
	) {
		iterate_circle_edge(
				center, r,
				[=](RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y) { painter->add_voxel(x, y, z, outline); }
		);
	}

	template<class voxel_painter_t = RendererApi::VirtualVoxelPainter>
	void draw_circle_edge(
			voxel_painter_t *painter,
			RendererApi::PointI<2>::component_t cx, RendererApi::PointI<2>::component_t cy,
			RendererApi::PointI<2>::component_t r,
			double z,
			RendererApi::Color::Transparent outline,
			RendererApi::SpriteMetadata *owner
	) {
		iterate_circle_edge(
				cx, cy, r,
				[=](RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y) { painter->add_voxel(x, y, z, outline, owner); }
		);
	}

	template<class voxel_painter_t = RendererApi::VirtualVoxelPainter>
	void draw_circle_edge(
			voxel_painter_t *painter,
			RendererApi::PointI<2> center,
			RendererApi::PointI<2>::component_t r,
			double z,
			RendererApi::Color::Transparent outline,
			RendererApi::SpriteMetadata *owner
	) {
		iterate_circle_edge(
				center, r,
				[=](RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y) { painter->add_voxel(x, y, z, outline, owner); }
		);
	}
}