#pragma once

#include <cstdlib>
#include <bgtu/computer_graphics_lab_work/renderer_api/point.hpp>

namespace BGTU::ComputerGraphicsLabWork::Utilities::Shapes {
	class brezenham_line_iterable {
	private:
		RendererApi::PointI<2>::component_t start_arg, start_ret, end_arg;
		RendererApi::PointI<2>::component_t delta_ret, delta_arg;
		RendererApi::PointI<2>::component_t delta_error, error_threshold;
		bool swap_flag;
	public:
		constexpr brezenham_line_iterable(
				RendererApi::PointI<2>::component_t x1,
				RendererApi::PointI<2>::component_t y1,
				RendererApi::PointI<2>::component_t x2,
				RendererApi::PointI<2>::component_t y2
		) noexcept {
			RendererApi::PointI<2>::component_t delta_x = (x1 > x2) ? (x1 - x2) : (x2 - x1);
			RendererApi::PointI<2>::component_t delta_y = (y1 > y2) ? (y1 - y2) : (y2 - y1);
			if (delta_x > delta_y) {
				this->start_arg = x1;
				this->start_ret = y1;
				this->end_arg = x2;
				this->delta_error = (RendererApi::PointI<2>::component_t) (delta_y + 1);
				this->error_threshold = (RendererApi::PointI<2>::component_t) (delta_x + 1);
				this->delta_ret = (y2 > y1) ? 1 : -1;
				this->swap_flag = false;
			} else {
				this->start_arg = y1;
				this->start_ret = x1;
				this->end_arg = y2;
				this->delta_error = (RendererApi::PointI<2>::component_t) (delta_x + 1);
				this->error_threshold = (RendererApi::PointI<2>::component_t) (delta_y + 1);
				this->delta_ret = (x2 > x1) ? 1 : -1;
				this->swap_flag = true;
			}
			this->delta_arg = (this->end_arg > this->start_arg) ? 1 : -1;
		}

		constexpr brezenham_line_iterable(
				RendererApi::PointI<2> start,
				RendererApi::PointI<2> end
		) noexcept: brezenham_line_iterable{start.x, start.y, end.x, end.y} {}

		class iterator {
		private:
			RendererApi::PointI<2>::component_t arg, ret, end_arg;
			RendererApi::PointI<2>::component_t delta_ret, delta_arg;
			RendererApi::PointI<2>::component_t error, delta_error, error_threshold;
			bool swap_flag;

			constexpr iterator(
					RendererApi::PointI<2>::component_t arg, RendererApi::PointI<2>::component_t ret, RendererApi::PointI<2>::component_t end_arg,
					RendererApi::PointI<2>::component_t delta_ret, RendererApi::PointI<2>::component_t delta_arg,
					RendererApi::PointI<2>::component_t delta_error, RendererApi::PointI<2>::component_t error_threshold,
					bool swap_flag
			) noexcept: arg{arg}, ret{ret}, end_arg{end_arg}, delta_ret{delta_ret}, delta_arg{delta_arg}, error{0}, delta_error{delta_error}, swap_flag{swap_flag}, error_threshold{error_threshold} {}

			friend class brezenham_line_iterable;

		public:
			constexpr iterator &operator++() noexcept {
				if (*this == nullptr)
					return *this;
				this->arg += this->delta_arg;
				this->error += this->delta_error;
				if (this->error >= this->error_threshold) {
					this->ret += this->delta_ret;
					this->error -= this->error_threshold;
				}
				return *this;
			}

			constexpr iterator operator++(int) noexcept {
				iterator next = *this;
				++(*this);
				return next;
			}

			constexpr RendererApi::PointI<2> operator*() const noexcept {
				if (this->swap_flag)
					return {this->ret, this->arg};
				else
					return {this->arg, this->ret};
			}

			constexpr bool operator==(std::nullptr_t) const noexcept {
				return (this->arg > this->end_arg) ^ (this->delta_arg < 0);
			}

			constexpr bool operator!=(std::nullptr_t) const noexcept {
				return !(*this == nullptr);
			}
		};

		[[nodiscard]] constexpr iterator begin() const noexcept {
			return iterator{
					this->start_arg, this->start_ret, this->end_arg,
					this->delta_ret, this->delta_arg,
					this->delta_error, this->error_threshold,
					this->swap_flag
			};
		}

		[[nodiscard]] constexpr std::nullptr_t end() const noexcept {
			return nullptr;
		}
	};
}