#pragma once

#include <cstdlib>
#include <bgtu/computer_graphics_lab_work/renderer_api/point.hpp>
#include <bgtu/computer_graphics_lab_work/renderer_api/voxel_painter.hpp>
#include <bgtu/computer_graphics_lab_work/renderer_api/sprite.hpp>
#include "../color.hpp"
#include "line.hpp"

namespace BGTU::ComputerGraphicsLabWork::Utilities::Shapes {
	template<class receiver_t, std::size_t points_count>
	void iterate_polyline(
			RendererApi::PointI<2> (&&points)[points_count],
			receiver_t receiver
	) {
		if constexpr (points_count == 0) {
			return;
		} else if constexpr (points_count == 1) {
			receiver(points[0].x, points[0].y);
			return;
		} else {
			receiver(points[0].x, points[0].y);
			for (std::size_t i = 1; i < points_count; i++) {
				if (points[i - 1] == points[i])
					continue;
				auto it = brezenham_line_iterable{points[i - 1], points[i]}.begin();
				it++;
				for (; it != nullptr; it++) {
					receiver((*it).x, (*it).y);
				}
			}
		}
	}


	template<std::size_t points_count, class voxel_painter_t = RendererApi::VirtualVoxelPainter>
	void draw_polyline_edge(
			voxel_painter_t *painter,
			RendererApi::PointI<2> (&&points)[points_count],
			double z,
			RendererApi::Color::Transparent outline
	) {
		iterate_polyline(
				points,
				[=](RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y) { painter->add_voxel(x, y, z, outline); }
		);
	}

	template<std::size_t points_count, class voxel_painter_t = RendererApi::VirtualVoxelPainter>
	void draw_polyline_edge(
			voxel_painter_t *painter,
			RendererApi::PointI<2> (&&points)[points_count],
			double z,
			RendererApi::Color::Transparent outline,
			RendererApi::SpriteMetadata *owner
	) {
		iterate_polyline(
				points,
				[=](RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y) { painter->add_voxel(x, y, z, outline, owner); }
		);
	}
}